import functools
from typing import TypeVar, Callable

T = TypeVar("T")


class WearyContext:
    def __init__(self) -> None:
        super(WearyContext, self).__init__()


def _decorate_with_context(property_name: str, method_impl: Callable) -> Callable:
    """
    Normally a function for a property has only the `self`
    as an attribute. Since we want to also pass some context
    to that function, we wrap the call in this function that
    creates the context, and passes it. The `decorated_function`
    still has only `self` as an argument.

    :param property_name:
    :param method_impl:
    :return:
    """

    @functools.wraps(method_impl)
    def decorated_function(self):
        if property_name in self._data:
            return self._data[property_name]

        ctx = WearyContext()

        result = method_impl(self, ctx)
        self._data[property_name] = result

        return result

    return decorated_function



import unittest
from typing import List
from unittest import TestCase

import weary

# generated
from weary import WearyContext


@weary.model
class TestEnvironment:
    @property
    def name(self) -> str:
        raise Exception("No provider for TestEnvironment.name")


# generated
@weary.model
class ApplicationModel:
    @property
    def versions(self) -> List[str]:
        raise Exception("No provider for ApplicationModel.versions")

    @property
    def environments(self) -> List[TestEnvironment]:
        raise Exception("No provider for ApplicationModel.environments")


call_count = 0


# user
@weary.property(ApplicationModel, "environments")
def resolve_environments_property(
    self: ApplicationModel, context: WearyContext
) -> List[TestEnvironment]:
    global call_count
    call_count += 1

    return []


# user
@weary.property(ApplicationModel, "versions")
def resolve_versions_property(
    self: ApplicationModel, context: WearyContext
) -> List[str]:
    return ["1", "2", "3"]


class TestWearyPropertyResolving(TestCase):
    def test_resolve_property(self):
        app = ApplicationModel()

        self.assertEqual(["1", "2", "3"], app.versions)

    def test_caching(self):
        app = ApplicationModel()
        self.assertEqual([], app.environments)
        self.assertEqual([], app.environments)
        self.assertEqual(1, call_count)

    def test_constructor_overrides_property(self):
        app = ApplicationModel(versions=["a", "b", "c"])
        self.assertEqual(["a", "b", "c"], app.versions)

    @unittest.expectedFailure
    def test_wrong_property_in_constructor_raises_exception(self):
        ApplicationModel(appversions=["a", "b", "c"])



import types
from typing import Callable, TypeVar

from weary.context import _decorate_with_context
from weary.method_registrations import method_registrations

T = TypeVar("T")


def property(clazz, property_name) -> Callable[..., Callable[..., T]]:
    if not hasattr(clazz, "_weary_base"):
        raise Exception(
            "You can only override properties for classes decorated "
            "with `@weary.model`."
        )

    def wrapper_builder(f: Callable[..., T]) -> Callable[..., T]:
        context_aware_function = _decorate_with_context(property_name, f)
        method_registrations[clazz._weary_base][property_name] = f
        property_definition = types.DynamicClassAttribute(
            context_aware_function, clazz._weary_base
        )
        setattr(
            clazz._weary_base,
            property_name,
            property_definition,
        )

        def set_value(self, value) -> None:
            self._data[property_name] = value

        property_definition.fset = set_value  # type: ignore

        return f

    return wrapper_builder



import functools
from typing import TypeVar, Type, Callable

from weary.method_registrations import method_registrations

T = TypeVar("T")


def model(f: Type[T]) -> Callable[..., T]:
    method_registrations[f] = dict()

    @functools.wraps(f)
    def wrapper(*args, **kw) -> T:
        return _create(f, *args, **kw)

    wrapper._weary_base = f  # type: ignore

    return wrapper


def _create(t: Type[T], *args, **kw) -> T:
    """
    Create the instance with the given type
    :param t:
    :return:
    """
    result = t()  # type: ignore
    result._data = dict()  # type: ignore

    if t not in method_registrations:
        raise Exception(f"{t} was not registered with @weary.model")

    for property_name, property_value in kw.items():
        if not hasattr(result, property_name):
            raise Exception(
                f"{property_name} is not defined on the @weary.model " f"class."
            )
        setattr(result, property_name, property_value)

    return result
